﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Threading;

namespace ProducerConsumerQ4
{
    class Program
    {
        static void Main(string[] args)
        {
            int limit = 10;
            var cq = new ConcurrentQueue<int>();
            List<Thread> consumers = new List<Thread>();
            SemaphoreSlim elementCounter = new SemaphoreSlim(0, limit);
            SemaphoreSlim upperLimit = new SemaphoreSlim(limit, limit);
            //the producer puts 100 elements into the collection at random intervals of up to 500ms
            var producer = new Thread(() =>
            {
                Random rand = new Random(Thread.CurrentThread.ManagedThreadId);
                for (int l = 0; l < 100; l++)
                {
                    upperLimit.Wait();
                    Console.WriteLine($"Put  {l}");
                    cq.Enqueue(l);
                    elementCounter.Release();
                    Thread.Sleep(rand.Next(500));
                }
            });
            producer.Start();
            Thread.Sleep(5000); //wait for hit the limit
            //Consumers take data from the collection at intervals of up to 4000 ms
            for (int i = 0; i < 10; i++)
            {
                var consumer = new Thread(() =>
                {
                    Random rand = new Random(Thread.CurrentThread.ManagedThreadId);
                    for (int l = 0; l < 10; l++)
                    {
                        Thread.Sleep(rand.Next(4000));
                        int result;
                        elementCounter.Wait();
                        bool succeeded = cq.TryDequeue(out result);
                        if (!succeeded)
                        {
                            throw new Exception("OMG we shouldn't be here");
                        }
                        upperLimit.Release();
                        Console.WriteLine($"Take {result} ");
                    }
                });
                consumers.Add(consumer);
                consumer.Start();
            }
            foreach (var th in consumers)
            {
                th.Join();
            }
            producer.Join();
            Console.WriteLine("That's all. Press Enter to finish");
            Console.ReadLine();
        }
    }
}
